# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from libc.stdint cimport int32_t, uint8_t


cdef extern from "python_krb5.h":
    ctypedef int32_t krb5_int32
    ctypedef krb5_int32 krb5_error_code
    ctypedef krb5_int32 krb5_deltat
    ctypedef krb5_error_code krb5_magic
    ctypedef krb5_int32 krb5_enctype
    ctypedef uint8_t krb5_octet
    ctypedef krb5_int32 krb5_timestamp
    ctypedef unsigned int krb5_boolean
    ctypedef krb5_int32 krb5_flags

    cdef struct _krb5_context:
        pass
    ctypedef _krb5_context *krb5_context

    cdef struct krb5_principal_data:
        pass
    ctypedef krb5_principal_data *krb5_principal
    ctypedef const krb5_principal_data *krb5_const_principal

    cdef struct _krb5_ccache:
        pass
    ctypedef _krb5_ccache *krb5_ccache

    cdef struct _krb5_get_init_creds_opt:
        pass
    ctypedef _krb5_get_init_creds_opt krb5_get_init_creds_opt

    cdef struct _krb5_init_creds_context:
        pass
    ctypedef _krb5_init_creds_context *krb5_init_creds_context

    cdef struct _krb5_keyblock:
        krb5_magic magic
        krb5_enctype enctype
        unsigned int length
        krb5_octet *contexts
    ctypedef _krb5_keyblock krb5_keyblock

    cdef struct _krb5_ticket_times:
        krb5_timestamp authtime
        krb5_timestamp starttime
        krb5_timestamp endtime
        krb5_timestamp renew_till
    ctypedef _krb5_ticket_times krb5_ticket_times

    cdef struct _krb5_address:
        pass
    ctypedef _krb5_address krb5_address

    cdef struct _krb5_data:
        krb5_magic magic
        unsigned int length
        char *data
    ctypedef _krb5_data krb5_data

    cdef struct _krb5_authdata:
        pass
    ctypedef _krb5_authdata krb5_authdata

    cdef struct _krb5_kt:
        pass
    ctypedef _krb5_kt *krb5_keytab

    cdef struct _krb5_creds:
        krb5_magic magic
        krb5_principal client
        krb5_principal server
        krb5_keyblock keyblock
        krb5_ticket_times times
        krb5_boolean is_skey
        krb5_flags ticket_flags
        krb5_address **addresses
        krb5_data ticket
        krb5_data second_ticket
        krb5_authdata **authdata
    ctypedef _krb5_creds krb5_creds

    cdef struct _krb5_prompt:
        char *prompt
        int hidden
        krb5_data *reply
    ctypedef _krb5_prompt krb5_prompt

    ctypedef krb5_error_code(*krb5_prompter_fct)(
        krb5_context context,
        void *data,
        const char *name,
        const char *banner,
        int num_prompts,
        krb5_prompt prompts[],
    )



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._context import Context

class KeyTab:
    """Kerberos KeyTab object.

    This class represents a Kerberos key table.

    Args:
        context: Krb5 context.
    """

    @property
    def addr(self) -> typing.Optional[int]:
        """The raw krb5_keytab pointer address of this credential cache."""
    @property
    def name(self) -> typing.Optional[bytes]:
        """The name/residual of the keytab."""
    @property
    def kt_type(self) -> typing.Optional[bytes]:
        """The type of the keytab."""

def kt_default(
    context: Context,
) -> KeyTab:
    """Resolve the default key table.

    Get a handle to the default keytab.

    Args:
        context: Krb5 context.

    Returns:
        KeyTab: The default keytab.
    """

def kt_default_name(
    context: Context,
) -> bytes:
    """Get the default key table name.

    Gets the name of the default key table for the context specified.

    Args:
        context: Krb5 context.

    Returns:
        bytes: The default key table name.
    """

def kt_get_name(
    context: Context,
    keytab: KeyTab,
) -> bytes:
    """Get a key table name.

    Get the name of the specified key table. See :meth:`kt_get_type()` to get
    the type of a keytab.

    Note:
        MIT Kerberos returns the full name whereas Heimdal just returns the
        name/residual. See :meth:`krb5_get_full_name` for the equivalent in
        Heimdal.

    Args:
        context: Krb5 context.
        keytab: The keytab to query.

    Returns:
        bytes: The name of the keytab.
    """

def kt_get_type(
    context: Context,
    keytab: KeyTab,
) -> bytes:
    """Get a key table type.

    Get the type of the specified key table. See :meth:`ky_get_name()` to get
    the name of a keytab.

    Args:
        context: Krb5 context.
        keytab: The keytab to query.

    Returns:
        bytes: The type of the keytab.
    """

def kt_resolve(
    context: Context,
    name: bytes,
) -> KeyTab:
    """Get a handle for a key table.

    Resolve the key table name and open a handle. The name must be of the from
    ``type:residual`` where type must be known to the library. If no type is
    specified then ``FILE`` is used as a default. The ``residual`` value is
    dependent on the type specified.

    Args:
        context: Krb5 context.
        name: The name of the keytab in the form ``type:residual``.

    Returns:
        KeyTab: The opened keytab.
    """



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._ccache import CCache
from krb5._context import Context
from krb5._creds_opt import GetInitCredsOpt

def get_init_creds_opt_set_out_ccache(
    context: Context,
    opt: GetInitCredsOpt,
    ccache: CCache,
) -> None:
    """Set output credential cache in options.

    Sets the output credential cache in the credential option structure.

    Args:
        context: Krb5 context.
        opt: The initial credential options.
        ccache: The credential cache to set as the output.
    """



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from libc.stdint cimport uintptr_t

from krb5._exceptions import Krb5Error
from krb5._principal import PrincipalParseFlags

from krb5._context cimport Context
from krb5._creds cimport Creds
from krb5._krb5_types cimport *
from krb5._principal cimport Principal


cdef extern from "python_krb5.h":
    krb5_error_code krb5_cc_cache_match(
        krb5_context context,
        krb5_principal client,
        krb5_ccache *cache_out,
    ) nogil

    krb5_error_code krb5_cc_close(
        krb5_context context,
        krb5_ccache cache,
    ) nogil

    krb5_error_code krb5_cc_default(
        krb5_context context,
        krb5_ccache *cache,
    ) nogil

    const char *krb5_cc_default_name(
        krb5_context context,
    ) nogil

    krb5_error_code krb5_cc_destroy(
        krb5_context context,
        krb5_ccache cache,
    ) nogil

    krb5_error_code krb5_cc_initialize(
        krb5_context context,
        krb5_ccache cache,
        krb5_principal principal,
    ) nogil

    const char *krb5_cc_get_name(
        krb5_context context,
        krb5_ccache cache,
    ) nogil

    krb5_error_code krb5_cc_get_principal(
        krb5_context context,
        krb5_ccache cache,
        krb5_principal *principal,
    ) nogil

    const char *krb5_cc_get_type(
        krb5_context context,
        krb5_ccache cache,
    ) nogil

    krb5_error_code krb5_cc_new_unique(
        krb5_context context,
        const char *type,
        const char *hint,
        krb5_ccache *id,
    ) nogil

    krb5_error_code krb5_cc_resolve(
        krb5_context context,
        const char *name,
        krb5_ccache *cache,
    ) nogil

    krb5_error_code krb5_cc_store_cred(
        krb5_context context,
        krb5_ccache cache,
        krb5_creds *creds,
    ) nogil

    krb5_boolean krb5_cc_support_switch(
        krb5_context context,
        const char *type,
    ) nogil

    krb5_error_code krb5_cc_switch(
        krb5_context context,
        krb5_ccache cache,
    ) nogil


cdef class CCache:
    # cdef Context ctx
    # cdef krb5_ccache raw

    def __cinit__(CCache self, Context context):
        self.ctx = context
        self.raw = NULL

    def __dealloc__(CCache self):
        if self.raw:
            krb5_cc_close(self.ctx.raw, self.raw)
            self.raw = NULL

    @property
    def addr(self) -> typing.Optional[int]:
        if self.raw:
            return <uintptr_t>self.raw

    @property
    def name(self) -> typing.Optional[bytes]:
        if self.raw:
            return cc_get_name(self.ctx, self)

    @property
    def principal(self) -> typing.Optional[Principal]:
        if self.raw:
            return cc_get_principal(self.ctx, self)

    @property
    def cache_type(self) -> typing.Optional[bytes]:
        if self.raw:
            return cc_get_type(self.ctx, self)

    def __repr__(CCache self) -> str:
        if self.raw:
            kwargs = [f"{k}={v}" for k, v in {
                'cache_type': self.cache_type.decode('utf-8'),
                'name': self.name.decode('utf-8'),
            }.items()]

            return f"CCache({', '.join(kwargs)})"

        else:
            return "CCache(NULL)"

    def __str__(CCache self) -> str:
        if self.raw:
            return f"{self.cache_type.decode('utf-8')}:{self.name.decode('utf-8')}"

        else:
            return "NULL"


def cc_cache_match(
    Context context not None,
    Principal principal not None,
) -> CCache:
    ccache = CCache(context)
    cdef krb5_error_code err = 0

    err = krb5_cc_cache_match(context.raw, principal.raw, &ccache.raw)
    if err:
        raise Krb5Error(context, err)

    return ccache


def cc_default(
    Context context not None,
) -> CCache:
    ccache = CCache(context)
    cdef krb5_error_code err = 0

    err = krb5_cc_default(context.raw, &ccache.raw)
    if err:
        raise Krb5Error(context, err)

    return ccache


def cc_default_name(
    Context context not None,
) -> bytes:
    return <bytes>krb5_cc_default_name(context.raw)


def cc_destroy(
    Context context not None,
    CCache cache not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_cc_destroy(context.raw, cache.raw)
    if err:
        raise Krb5Error(context, err)
    cache.raw = NULL  # Stops dealloc from calling close


def cc_get_name(
    Context context not None,
    CCache cache not None,
) -> bytes:
    return krb5_cc_get_name(context.raw, cache.raw)


def cc_get_principal(
    Context context not None,
    CCache cache not None,
) -> Principal:
    princ = Principal(context, PrincipalParseFlags.none)
    cdef krb5_error_code err = 0

    err = krb5_cc_get_principal(context.raw, cache.raw, &princ.raw)
    if err:
        raise Krb5Error(context, err)

    return princ


def cc_get_type(
    Context context not None,
    CCache cache not None,
) -> bytes:
    return krb5_cc_get_type(context.raw, cache.raw)


def cc_initialize(
    Context context not None,
    CCache cache not None,
    Principal principal not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_cc_initialize(context.raw, cache.raw, principal.raw)
    if err:
        raise Krb5Error(context, err)


def cc_new_unique(
    Context context not None,
    const unsigned char[:] cred_type not None,
    const unsigned char[:] hint = None,
) -> CCache:
    ccache = CCache(context)
    cdef krb5_error_code err = 0

    # TODO: Test that this cannot be NULL
    if not len(cred_type):
        raise ValueError("cred_type must be set to a valid type value not an empty byte string")

    cdef const char *hint_ptr = NULL
    if hint is not None and len(hint):
        hint_ptr = <const char*>&hint[0]

    err = krb5_cc_new_unique(context.raw, <const char*>&cred_type[0], hint_ptr, &ccache.raw)
    if err:
        raise Krb5Error(context, err)

    return ccache


def cc_resolve(
    Context context not None,
    const unsigned char[:] name not None,
) -> CCache:
    ccache = CCache(context)
    cdef krb5_error_code err = 0

    cdef const char *name_ptr = NULL
    if len(name):
        name_ptr = <const char*>&name[0]
    else:
        raise ValueError("name cannot be an empty byte string")

    err = krb5_cc_resolve(context.raw, name_ptr, &ccache.raw)
    if err:
        raise Krb5Error(context, err)

    return ccache


def cc_store_cred(
    Context context not None,
    CCache cache not None,
    Creds creds not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_cc_store_cred(context.raw, cache.raw, &creds.raw)
    if err:
        raise Krb5Error(context, err)


def cc_support_switch(
    Context context not None,
    const unsigned char[:] cache_type not None,
) -> bool:
    cdef const char *type_ptr = NULL
    if len(cache_type):
        type_ptr = <const char*>&cache_type[0]
    else:
        raise ValueError("cache_type cannot be an empty byte string")

    return bool(krb5_cc_support_switch(context.raw, type_ptr))


def cc_switch(
    Context context not None,
    CCache cache not None,
) -> None:
    cdef krb5_error_code err = 0

    err = krb5_cc_switch(context.raw, cache.raw)
    if err:
        raise Krb5Error(context, err)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

from krb5._ccache import (
    CCache,
    cc_cache_match,
    cc_default,
    cc_default_name,
    cc_destroy,
    cc_get_name,
    cc_get_principal,
    cc_get_type,
    cc_initialize,
    cc_new_unique,
    cc_resolve,
    cc_store_cred,
    cc_support_switch,
    cc_switch,
)
from krb5._context import Context, get_default_realm, init_context, set_default_realm
from krb5._creds import (
    Creds,
    InitCredsContext,
    Krb5Prompt,
    get_init_creds_keytab,
    get_init_creds_password,
    init_creds_get,
    init_creds_get_creds,
    init_creds_init,
    init_creds_set_keytab,
    init_creds_set_password,
)
from krb5._creds_opt import (
    GetInitCredsOpt,
    get_init_creds_opt_alloc,
    get_init_creds_opt_set_canonicalize,
    get_init_creds_opt_set_forwardable,
)
from krb5._exceptions import Krb5Error
from krb5._kt import (
    KeyTab,
    kt_default,
    kt_default_name,
    kt_get_name,
    kt_get_type,
    kt_resolve,
)
from krb5._principal import (
    Principal,
    PrincipalParseFlags,
    PrincipalUnparseFlags,
    parse_name_flags,
    unparse_name_flags,
)

__all__ = [
    "CCache",
    "Context",
    "Creds",
    "GetInitCredsOpt",
    "InitCredsContext",
    "KeyTab",
    "Krb5Error",
    "Krb5Prompt",
    "Principal",
    "PrincipalParseFlags",
    "PrincipalUnparseFlags",
    "cc_cache_match",
    "cc_default",
    "cc_default_name",
    "cc_destroy",
    "cc_get_name",
    "cc_get_principal",
    "cc_get_type",
    "cc_initialize",
    "cc_new_unique",
    "cc_resolve",
    "cc_store_cred",
    "cc_support_switch",
    "cc_switch",
    "get_default_realm",
    "get_init_creds_keytab",
    "get_init_creds_opt_alloc",
    "get_init_creds_opt_set_canonicalize",
    "get_init_creds_opt_set_forwardable",
    "get_init_creds_password",
    "init_context",
    "init_creds_get",
    "init_creds_get_creds",
    "init_creds_init",
    "init_creds_set_keytab",
    "init_creds_set_password",
    "kt_default",
    "kt_default_name",
    "kt_get_name",
    "kt_get_type",
    "kt_resolve",
    "parse_name_flags",
    "set_default_realm",
    "unparse_name_flags",
]

# Provider specific APIs
try:
    from krb5._ccache_mit import cc_dup
except ImportError:
    pass
else:
    __all__.append("cc_dup")


try:
    from krb5._context_mit import init_secure_context
except ImportError:
    pass
else:
    __all__.append("init_secure_context")


try:
    from krb5._creds_opt_heimdal import get_init_creds_opt_set_default_flags
except ImportError:
    pass
else:
    __all__.append("get_init_creds_opt_set_default_flags")


try:
    from krb5._creds_opt_mit import get_init_creds_opt_set_out_ccache
except ImportError:
    pass
else:
    __all__.append("get_init_creds_opt_set_out_ccache")


try:
    from krb5._kt_mit import kt_client_default, kt_dup
except ImportError:
    pass
else:

    def kt_get_full_name(
        context: Context,
        keytab: KeyTab,
    ) -> bytes:
        # This isn't implemented in MIT so mock the same behaviour
        return (keytab.kt_type or b"") + b":" + (keytab.name or b"")

    __all__.extend(["kt_client_default", "kt_dup", "kt_get_full_name", "kt_get_type"])


try:
    from krb5._kt_heimdal import kt_get_full_name
except ImportError:
    pass
else:
    __all__.extend(["kt_get_full_name", "kt_get_type"])


try:
    from krb5._principal_heimdal import principal_get_realm
except ImportError:
    pass
else:
    __all__.extend(["principal_get_realm"])



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import typing

from krb5._context import Context

class GetInitCredsOpt:
    """Kerberos Initial Credential Options object.

    This class represents an initial credential options object.

    Args:
        context: Krb5 context.
    """

def get_init_creds_opt_alloc(
    context: Context,
) -> GetInitCredsOpt:
    """Allocate a new initial credential options object.

    Creates a new options structure that control how credential are gotten.

    Args:
        context: Krb5 context.

    Returns:
        GetInitCredsOpt: The initial credential options object.
    """

def get_init_creds_opt_set_canonicalize(
    opt: GetInitCredsOpt,
    canonicalize: typing.Optional[bool],
) -> None:
    """Set canonicalization details in the initial credential options.

    Sets whether the credential client principal should be canonicalized by
    the KDC.

    Args:
        opt: The initial credential options.
        canonicalize: Whether to set or unset the canonicalize option.
    """

def get_init_creds_opt_set_forwardable(
    opt: GetInitCredsOpt,
    forwardable: typing.Optional[bool],
) -> None:
    """Set forwardable details in the initial credential options.

    Sets whether the credentials are to be forwardable or not.

    Args:
        opt: The initial credential options.
        forwardable: Whether to set or unset the forwardable option.
    """



#!/usr/bin/env python

# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

"""Builds the project.

Logic has mostly been derived from python-gssapi - thanks to them for all the
hard work.

https://github.com/pythongssapi/python-gssapi/blob/main/setup.py
"""

import ctypes
import ctypes.util
import glob
import os
import os.path
import platform
import shlex
import subprocess
import sys
import typing

from setuptools import Extension, find_packages, setup
from setuptools.command.sdist import sdist

SKIP_CYTHON_FILE = "__dont_use_cython__.txt"
SKIP_MODULE_CHECK = os.environ.get("KRB5_SKIP_MODULE_CHECK", "false").lower() == "true"
CYTHON_LINETRACE = os.environ.get("KRB5_CYTHON_TRACING", "false").lower() == "true"

if os.path.exists(SKIP_CYTHON_FILE):
    print("In distributed package, building from C files...")
    SOURCE_EXT = "c"
else:
    try:
        from Cython.Build import cythonize

        print("Building from Cython files...")
        SOURCE_EXT = "pyx"
    except ImportError:
        print("Cython not found, building from C files...")
        SOURCE_EXT = "c"


def run_command(*args: str) -> str:
    stdout = subprocess.check_output(args, shell=True)
    return stdout.decode("utf-8").strip()


def make_extension(
    name: str,
    module: ctypes.CDLL,
    canary: typing.Optional[str] = None,
    **kwargs: typing.Any,
) -> Extension:
    source = os.path.join("src", name.replace(".", os.sep)) + f".{SOURCE_EXT}"

    if not SKIP_MODULE_CHECK and canary and not hasattr(module, canary):
        print(f"Skipping {source} as it is not supported by the selected Kerberos implementation.")
        return

    if not os.path.exists(source):
        raise FileNotFoundError(source)

    print(f"Compiling {source}")
    return Extension(
        name=name,
        sources=[source],
        **kwargs,
    )


def get_krb5_config() -> str:
    """Gets the path to the krb5-config binary.

    Determines the path to the krb5-confing to use for detecting linker and
    compiler args on the current platform. The default behaviour is to use
    the value of the env var ``KRB5_KRB5CONFIG`` defaulting to just
    krb5-conf in the path.

    FreeBSD is slightly different as it favours ``/usr/local/bin/krb5-config``
    over ``/usr/bin/krb5-config`` due to the latter being based on an ancient
    Heimdal version.

    Returns:
        str: Path to krb5-conf.
    """
    kc_env = os.environ.get("KRB5_KRB5CONFIG", "")
    kc = kc_env if kc_env else "krb5-config"

    if sys.platform.startswith("freebsd") and not kc_env:
        # FreeBSD does $PATH backward, for our purposes.  That is, the package
        # manager's version of the software is in /usr/local, which is in PATH
        # *after* the version in /usr.  We prefer the package manager's version
        # because the Heimdal in base is truly ancient, but this can be overridden
        # - either in the "normal" fashion by putting something in PATH in front
        # of it, or by removing /usr/local from PATH.
        bins = []
        for entry in os.environ.get("PATH", "").split(os.pathsep):
            p = os.path.join(entry, "krb5-config")
            if not os.path.exists(p):
                continue
            bins.append(p)

        if len(bins) > 1 and bins[0] == "/usr/bin/krb5-config" and "/usr/local/bin/krb5-config" in bins:
            kc = "/usr/local/bin/krb5-config"

    return kc


def get_krb5_lib_path(
    libraries: typing.List[str],
    library_dirs: typing.List[str],
    link_args: typing.List[str],
    macos_native: bool,
) -> str:
    """Gets the path to the main libkrb5.so library.

    This path is used to check if the Kerberos library supports the optional
    APIs when compiling.

    Args:
        libraries: The libraries specified by krb5-config (-l).
        librariy_dirs: The library dirs specified by krb5-config (-L).
        link_args: The linking args specified by krb5-config.
        macos_native: Whether this is for the native macOS Kerberos library.

    Returns:
        str: The path to the krb5 library.
    """
    krb5_lib = os.environ.get("KRB5_MAIN_LIB", "")
    krb5_path = ""

    if not krb5_lib and macos_native:
        # Cannot use find_library as it won't look into the private framework path.
        krb5_lib = "/System/Library/PrivateFrameworks/Heimdal.framework/Heimdal"

    elif not krb5_lib:
        for opt in libraries:
            if opt.startswith("krb5"):
                ext = {
                    "nt": "dll",
                    "darwin": "dylib",
                }.get(os.name, "so")

                krb5_lib = f"lib{opt}.{ext}"
                break

        for opt in link_args:
            # To support Heimdal on Debian, read the linker path.
            if opt.startswith("-Wl,/"):
                krb5_path = opt[4:] + "/"
                break

        if not krb5_path:
            for d in library_dirs:
                if os.path.exists(os.path.join(d, krb5_lib)):
                    krb5_path = d
                    break

    if not krb5_lib:
        raise Exception(
            "Could not find the main krb5 shared library. Set this manually with the env var KRB5_MAIN_LIB."
        )

    return os.path.join(krb5_path, krb5_lib)


with open(os.path.join(os.path.dirname(__file__), "README.md"), mode="rb") as fd:
    long_description = fd.read().decode("utf-8")

kc = get_krb5_config()
print(f"Using krb5-config at '{kc}'")

macos_native = False
if sys.platform == "darwin":
    mac_ver = [int(v) for v in platform.mac_ver()[0].split(".")]
    macos_native = mac_ver >= [10, 7, 0]

compile_args, raw_link_args = [
    shlex.split(os.environ[e], posix=True) if e in os.environ else None
    for e in ["KRB5_COMPILER_ARGS", "KRB5_LINKER_ARGS"]
]
if compile_args is None:
    if macos_native:
        compile_args = []

    else:
        compile_args = shlex.split(run_command(f"{kc} --cflags krb5"))

    compile_args.append("-Werror")

    # Python 3.8 on macOS errors on these deprecation warnings. We ignore them as things are fixed on 3.9 but the
    # code still needs to compile on 3.8.
    if sys.platform == "darwin" and sys.version_info[:2] == (3, 8):
        compile_args.append("-Wno-deprecated")

    if CYTHON_LINETRACE:
        compile_args.append("-DCYTHON_TRACE_NOGIL=1")

if raw_link_args is None:
    if macos_native:
        raw_link_args = ["-framework", "Heimdal", "-F", "/System/Library/PrivateFrameworks"]

    else:
        raw_link_args = shlex.split(run_command(f"{kc} --libs krb5"))

library_dirs, include_dirs, libraries, link_args = [], [], [], []
for arg in raw_link_args:
    if arg.startswith("-L"):
        library_dirs.append(arg[2:])
    elif arg.startswith("-l"):
        libraries.append(arg[2:])
    elif arg.startswith("-I"):
        include_dirs.append(arg[2:])
    else:
        link_args.append(arg)

if macos_native:
    # Because this is doing a naughty thing and linking against a private framework there are no krb5.h header files
    # available. A hack is to just compile Heimdal itself and include that during compilation.
    heimdal_dir = os.environ.get("KRB5_MACOS_HEIMDAL_DIR", "")
    if not heimdal_dir:
        heimdal_dir = os.path.join(os.path.dirname(__file__), "build_helpers", "heimdal")

    include_dirs.append(os.path.join(os.path.abspath(heimdal_dir), "include"))


krb5_path = get_krb5_lib_path(libraries, library_dirs, link_args, macos_native)
print(f"Using {krb5_path} as Kerberos module for platform checks")
krb5 = ctypes.CDLL(krb5_path)

if hasattr(krb5, "krb5_xfree"):
    compile_args.append("-DHEIMDAL_XFREE")


raw_extensions = []
for e in [
    "ccache",
    ("ccache_mit", "krb5_cc_dup"),
    "context",
    ("context_mit", "krb5_init_secure_context"),
    "creds",
    "creds_opt",
    ("creds_opt_heimdal", "krb5_get_init_creds_opt_set_default_flags"),
    ("creds_opt_mit", "krb5_get_init_creds_opt_set_out_ccache"),
    "exceptions",
    "kt",
    ("kt_mit", "krb5_kt_dup"),
    ("kt_heimdal", "krb5_kt_get_full_name"),
    "principal",
    ("principal_heimdal", "krb5_principal_get_realm"),
]:
    name = e
    canary = None
    if isinstance(e, tuple):
        name = e[0]
        if len(e) > 1:
            canary = e[1]

    ext = make_extension(
        f"krb5._{name}",
        module=krb5,
        canary=canary,
        extra_link_args=link_args,
        extra_compile_args=compile_args,
        include_dirs=include_dirs,
        library_dirs=library_dirs,
        libraries=libraries,
    )
    if ext:
        raw_extensions.append(ext)

if SOURCE_EXT == "c":
    extensions = raw_extensions
else:
    extensions = cythonize(
        raw_extensions,
        language_level=3,
        compiler_directives={"linetrace": CYTHON_LINETRACE},
    )


class sdist_krb5(sdist):
    def run(self) -> None:
        if not self.dry_run:
            with open(SKIP_CYTHON_FILE, mode="wb") as flag_file:
                flag_file.write(b"")

            sdist.run(self)

            os.remove(SKIP_CYTHON_FILE)


setup(
    name="krb5",
    version="0.2.0",
    packages=find_packages(where="src"),
    package_data={
        "krb5": ["py.typed", "*.pyi"],
    },
    package_dir={"": "src"},
    py_modules=[os.path.splitext(os.path.basename(path))[0] for path in glob.glob("src/*.py")],
    cmdclass={
        "sdist": sdist_krb5,
    },
    ext_modules=extensions,
    include_package_data=False,
    install_requires=[],
    extras_require={},
    author="Jordan Borean",
    author_email="jborean93@gmail.com",
    url="https://github.com/jborean93/pykrb5",
    description="Kerberos API bindings for Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="krb5 kerberos",
    license="MIT",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import pathlib

import k5test
import pytest

import krb5


def test_kt_default(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_default(ctx)
    assert isinstance(kt, krb5.KeyTab)
    assert isinstance(kt.addr, int)
    assert kt.name == realm.keytab.encode()
    assert kt.kt_type == b"FILE"

    assert repr(kt) == f"KeyTab(kt_type=FILE, name={realm.keytab})"
    assert str(kt) == f"FILE:{realm.keytab}"


def test_kt_default_name(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()

    actual = krb5.kt_default_name(ctx)
    assert actual == realm.keytab.encode()


def test_kt_get_name(realm: k5test.K5Realm, tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    actual = krb5.kt_get_name(ctx, kt)

    if realm.provider == "mit":
        assert actual == f"FILE:{tmp_path / 'keytab'}".encode()
    else:
        assert actual == f"{tmp_path / 'keytab'}".encode()


def test_kt_get_full_name(tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    assert krb5.kt_get_full_name(ctx, kt) == f"FILE:{tmp_path / 'keytab'}".encode()


def test_kt_get_type(tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    assert krb5.kt_get_type(ctx, kt) == b"FILE"


def test_kt_resolve(tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    assert isinstance(kt, krb5.KeyTab)
    assert kt.kt_type == b"FILE"
    assert kt.name == f"{tmp_path / 'keytab'}".encode()
    assert str(kt) == f"FILE:{tmp_path / 'keytab'}"
    assert repr(kt) == f"KeyTab(kt_type=FILE, name={tmp_path / 'keytab'})"


@pytest.mark.requires_api("kt_client_default")
def test_kt_client_default(realm: k5test.K5Realm) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_client_default(ctx)
    assert isinstance(kt, krb5.KeyTab)
    assert isinstance(kt.addr, int)
    assert kt.name == realm.client_keytab.encode()
    assert kt.kt_type == b"FILE"

    assert repr(kt) == f"KeyTab(kt_type=FILE, name={realm.client_keytab})"
    assert str(kt) == f"FILE:{realm.client_keytab}"


@pytest.mark.requires_api("kt_dup")
def test_kt_dup(tmp_path: pathlib.Path) -> None:
    ctx = krb5.init_context()

    kt = krb5.kt_resolve(ctx, f"FILE:{tmp_path / 'keytab'}".encode())
    kt_name = kt.name
    kt_addr = kt.addr

    copied_kt = krb5.kt_dup(ctx, kt)
    del kt
    assert isinstance(copied_kt, krb5.KeyTab)
    assert copied_kt.kt_type == b"FILE"
    assert copied_kt.name == kt_name
    assert copied_kt.addr != kt_addr



# Copyright: (c) 2021 Jordan Borean (@jborean93) <jborean93@gmail.com>
# MIT License (see LICENSE or https://opensource.org/licenses/MIT)

import os

import k5test
import pytest

import krb5


def test_get_init_creds_opt_alloc() -> None:
    ctx = krb5.init_context()

    opt = krb5.get_init_creds_opt_alloc(ctx)
    assert isinstance(opt, krb5.GetInitCredsOpt)
    assert str(opt) == "GetInitCredsOpt"


def test_get_init_creds_opt_set_canonicalize() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_canonicalize(opt, True)
    krb5.get_init_creds_opt_set_canonicalize(opt, False)
    krb5.get_init_creds_opt_set_canonicalize(opt, None)


def test_get_init_creds_opt_set_forwardable() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_forwardable(opt, True)
    krb5.get_init_creds_opt_set_forwardable(opt, False)
    krb5.get_init_creds_opt_set_forwardable(opt, None)


@pytest.mark.requires_api("get_init_creds_opt_set_out_ccache")
def test_get_init_creds_opt_set_out_ccache() -> None:
    ctx = krb5.init_context()
    ccache = krb5.cc_new_unique(ctx, b"MEMORY")

    opt = krb5.get_init_creds_opt_alloc(ctx)
    krb5.get_init_creds_opt_set_out_ccache(ctx, opt, ccache)


@pytest.mark.requires_api("get_init_creds_opt_set_default_flags")
def test_get_init_creds_opt_set_default_flags() -> None:
    ctx = krb5.init_context()
    opt = krb5.get_init_creds_opt_alloc(ctx)

    krb5.get_init_creds_opt_set_default_flags(ctx, opt)
    krb5.get_init_creds_opt_set_default_flags(ctx, opt, b"appname", b"realm")
